#include <stdlib.h>
#include <errno.h>
#include <locale.h>

#include "pc6001v.h"
#include "typedef.h"
#include "config.h"
#include "console.h"
#include "error.h"
#include "osd.h"
#include "p6el.h"
#include "p6vm.h"

#include "icons.h"



///////////////////////////////////////////////////////////
// tHgt@C`FbN(΍쐬)
///////////////////////////////////////////////////////////
bool CheckFont( void )
{
	char FontFile[PATH_MAX];
	bool ret = true;
	
	sprintf( FontFile, "%s%s/%s", OSD_GetModulePath(), FONT_DIR, FONTH_FILE );
	ret |= ( !OSD_FileExist( FontFile ) && !OSD_CreateFont( FontFile, NULL, FSIZE ) );
	
	sprintf( FontFile, "%s%s/%s", OSD_GetModulePath(), FONT_DIR, FONTZ_FILE );
	ret |= ( !OSD_FileExist( FontFile ) && !OSD_CreateFont( NULL, FontFile, FSIZE ) );
	
	return ret;
}


///////////////////////////////////////////////////////////
// ROMt@C݃`FbN&@ύX
///////////////////////////////////////////////////////////
bool SerchRom( cConfig *cfg )
{
	char RomSerch[PATH_MAX];
	
	int IniModel = cfg->GetModel();
	sprintf( RomSerch, "%s*.%2d", cfg->GetRomPath(), IniModel );
	if( OSD_FileExist( RomSerch ) ){
		Error::SetError( Error::NoError );
		return true;
	}
	
	int models[] = { 60, 62, 66 };
	for( int i=0; i < COUNTOF(models); i++ ){
		sprintf( RomSerch, "%s*.%2d", cfg->GetRomPath(), models[i] );
		if( OSD_FileExist( RomSerch ) ){
			cfg->SetModel( models[i] );
			Error::SetError( Error::RomChange );
			return true;
		}
	}
	Error::SetError( Error::NoRom );
	return false;
}


///////////////////////////////////////////////////////////
// ACRݒ
///////////////////////////////////////////////////////////
void SetIcon( int model )
{
	BYTE *ipix, *imsk;
	switch( model ){
	case 62: ipix = (BYTE *)p62pix; imsk = (BYTE *)p62msk; break;
	case 66: ipix = (BYTE *)p66pix; imsk = (BYTE *)p66msk; break;
	default: ipix = (BYTE *)p60pix; imsk = (BYTE *)p60msk;
	}
	SDL_Surface *p6icon = SDL_CreateRGBSurfaceFrom( ipix, 32, 32, 24, 32*3, 0, 0, 0, 0 );
	SDL_WM_SetIcon( p6icon, imsk );
	SDL_FreeSurface( p6icon );
}


///////////////////////////////////////////////////////////
// C
///////////////////////////////////////////////////////////
int main( int argc, char *argv[] )
{
	EL6 *P6Core             = NULL;			// IuWFNg|C^
	EL6::ReturnCode Restart = EL6::Quit;	// ċNtO
	cConfig Cfg;							// ݒIuWFNg
	
	
	// ϐݒ(eXgp)
//	putenv( "SDL_VIDEODRIVER=windib" );
//	putenv( "SDL_VIDEODRIVER=directx" );
//	putenv( "SDL_AUDIODRIVER=dsound" );
	
	setlocale( LC_CTYPE, "" );
	
	// dN֎~
	if( OSD_IsWorking() ) return false;
	
	// SDL
	if( SDL_Init( SDL_INIT_TIMER | SDL_INIT_AUDIO | SDL_INIT_VIDEO | SDL_INIT_JOYSTICK ) ){
		Error::SetError( Error::LibInitFailed );
		OSD_Message( (char *)Error::GetErrorText(), MSERR_ERROR, OSDR_OK | OSDM_ICONERROR );
		OSD_Finish();	// I
		return false;
	}
	
	// tHgt@C`FbN
	if( !CheckFont() ){
		Error::SetError( Error::FontCreateFailed );
		OSD_Message( (char *)Error::GetErrorText(), MSERR_ERROR, OSDR_OK | OSDM_ICONWARNING );
		Error::SetError( Error::NoError );
	}
	
	// R\[ptHgǍ
	char FontZ[PATH_MAX], FontH[PATH_MAX];
	sprintf( FontZ, "%s%s/%s", OSD_GetModulePath(), FONT_DIR, FONTZ_FILE );
	sprintf( FontH, "%s%s/%s", OSD_GetModulePath(), FONT_DIR, FONTH_FILE );
	if( !JFont::OpenFont( FontZ, FontH ) ){
		Error::SetError( Error::FontLoadFailed );
		OSD_Message( (char *)Error::GetErrorText(), MSERR_ERROR, OSDR_OK | OSDM_ICONERROR );
		Error::SetError( Error::NoError );
	}
	
	
	// INIt@CǍ
	if( !Cfg.Init() ){
		switch( Error::GetError() ){
		case Error::IniDefault:
			OSD_Message( (char *)Error::GetErrorText(), MSERR_ERROR, OSDR_OK | OSDM_ICONWARNING );
			Error::SetError( Error::NoError );
			break;
			
		default:
			OSD_Message( (char *)Error::GetErrorText(), MSERR_ERROR, OSDR_OK | OSDM_ICONERROR );
			SDL_Quit();				// SDLVbg_E
			OSD_Finish();			// I
			return false;
		}
	}
	
	
	// P6IuWFNg쐬Ďs
	do{
		// ROMt@C݃`FbN&@ύX
		if( SerchRom( &Cfg ) ){
			if( Error::GetError() != Error::NoError ){
				OSD_Message( (char *)Error::GetErrorText(), MSERR_ERROR, OSDR_OK | OSDM_ICONWARNING );
				Error::SetError( Error::NoError );
			}
		}else{
			OSD_Message( (char *)Error::GetErrorText(), MSERR_ERROR, OSDR_OK | OSDM_ICONERROR );
			break;	// do 
		}
		
		// ACRݒ
		SetIcon( Cfg.GetModel() );
		
		
		// @P6IuWFNgm
		P6Core = new EL6;
		if( !P6Core ){
			break;	// do 
		}
		
		// VM
		if( !P6Core->Init( &Cfg ) ){
			// sꍇ
			OSD_Message( (char *)Error::GetErrorText(), MSERR_ERROR, OSDR_OK | OSDM_ICONERROR );
			break;	// do 
		}
		
		switch( Restart ){
		case EL6::Dokoload:	// ǂłLOAD?
			P6Core->DokoDemoLoad( Cfg.GetDokoFile() );
			break;
			
		case EL6::Replay:	// vCĐ?
		{
			char temp[PATH_MAX];
			strncpy( temp, Cfg.GetDokoFile(), PATH_MAX );
			P6Core->DokoDemoLoad( temp );
			P6Core->REPLAY::StartReplay( temp );
		}
			break;
			
		default:
			break;
		}
		
		// VMs
		P6Core->Start();
		Restart = P6Core->EventLoop();
		P6Core->Stop();
		
		delete P6Core;	// P6IuWFNgJ
		
		
		// ċNȂINIt@CēǍ
		if( Restart == EL6::Restart ){
			if( !Cfg.Init() ){
				Error::SetError( Error::IniDefault );
				OSD_Message( (char *)Error::GetErrorText(), MSERR_ERROR, OSDR_OK | OSDM_ICONWARNING );
				Error::SetError( Error::NoError );
			}
		}
		
	}while( Restart != EL6::Quit );
	
	
	
	// SDLVbg_E
	SDL_Quit();
	
	// I
	OSD_Finish();
	
	return true;
}
